/*
 * Raspberry Pi PIC Programmer using GPIO connector
 * https://github.com/WallaceIT/picberry
 * Copyright 2016 Francesco Valla
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <vector>

#include "../common.h"
#include "device.h"

using namespace std;

#define SF_PIC32MX1		0x00
#define SF_PIC32MX2		0x01
#define SF_PIC32MX3		0x02
#define SF_PIC32MZ		0x03
#define SF_PIC32MK		0x04

extern vector<uint32_t> pe_loader;
extern vector<uint32_t> pic32_pemx1;
extern vector<uint32_t> pic32_pemx3;
extern vector<uint32_t> pic32_pemz;

class pic32: public Pic{

	public:
		pic32(uint8_t sf){
			subfamily=sf;
		};
		void enter_program_mode(void);
		void exit_program_mode(void);
		bool setup_pe(void);
		bool read_device_id(void);
		void bulk_erase(void);
		void dump_configuration_registers(void);
		void read(char *outfile, uint32_t start=0, uint32_t count=0);
		void write(char *infile);
		uint8_t blank_check(void);

	protected:
		uint8_t Data4Phase(uint8_t tdi, uint8_t tms);
		void Data2Phase(uint8_t tdi, uint8_t tms);
		void SetMode(uint8_t length, uint8_t mode);
		void SendCommand(uint8_t command);
		uint32_t XferData(uint8_t length, uint32_t iData);
		void XferFastData2P(uint32_t iData);
		uint32_t XferFastData4P(uint32_t iData);
		void XferInstruction(uint32_t instruction);
		uint32_t ReadFromAddress(uint32_t address);
		uint32_t GetPEResponse(void);
		bool check_device_status(void);
		void code_protected_bulk_erase(void);
		bool enter_serial_exec_mode(void);
		void download_pe(vector<uint32_t> pe_pointer);
		
		uint32_t bootsize;
		uint32_t rowsize;

		/*
		* DEVICES SECTION
		* 						 	ID       	NAME			MEMSIZE (16bit words)
		*/
		pic_device piclist[221] = { {0x0938053, "PIC32MX360F512L", 0x40000},
									{0x0934053, "PIC32MX360F256L", 0x20000},
									{0x092D053, "PIC32MX340F128L", 0x10000},
									{0x092A053, "PIC32MX320F128L", 0x10000},
									{0x0916053, "PIC32MX340F512H", 0x40000},
									{0x0912053, "PIC32MX340F256H", 0x20000},
									{0x090D053, "PIC32MX340F128H", 0x10000},
									{0x090A053, "PIC32MX320F128H", 0x10000},
									{0x0906053, "PIC32MX320F064H", 0xF000},
									{0x0978053, "PIC32MX460F512L", 0x40000},
									{0x0974053, "PIC32MX460F256L", 0x20000},
									{0x096D053, "PIC32MX440F128L", 0x10000},
									{0x0952053, "PIC32MX440F256H", 0x20000},
									{0x0956053, "PIC32MX440F512H", 0x40000},
									{0x094D053, "PIC32MX440F128H", 0x10000},
									{0x04317053, "PIC32MX575F256H", 0x20000},
									{0x0430B053, "PIC32MX675F256H", 0x20000},
									{0x04303053, "PIC32MX775F256H", 0x20000},
									{0x04309053, "PIC32MX575F512H", 0x40000},
									{0x0430C053, "PIC32MX675F512H", 0x40000},
									{0x04325053, "PIC32MX695F512H", 0x40000},
									{0x0430D053, "PIC32MX775F512H", 0x40000},
									{0x0430E053, "PIC32MX795F512H", 0x40000},
									{0x04333053, "PIC32MX575F256L", 0x20000},
									{0x04305053, "PIC32MX675F256L", 0x20000},
									{0x04312053, "PIC32MX775F256L", 0x20000},
									{0x0430F053, "PIC32MX575F512L", 0x40000},
									{0x04311053, "PIC32MX675F512L", 0x40000},
									{0x04341053, "PIC32MX695F512L", 0x40000},
									{0x04307053, "PIC32MX775F512L", 0x40000},
									{0x04307053, "PIC32MX795F512L", 0x40000},
									{0x04400053, "PIC32MX534F064H", 0xF000},
									{0x04401053, "PIC32MX564F064H", 0xF000},
									{0x04403053, "PIC32MX564F128H", 0x10000},
									{0x04405053, "PIC32MX664F064H", 0xF000},
									{0x04407053, "PIC32MX664F128H", 0x10000},
									{0x0440B053, "PIC32MX764F128H", 0x10000},
									{0x0440C053, "PIC32MX534F064L", 0xF000},
									{0x0440D053, "PIC32MX564F064L", 0xF000},
									{0x0440F053, "PIC32MX564F128L", 0x10000},
									{0x04411053, "PIC32MX664F064L", 0xF000},
									{0x04413053, "PIC32MX664F128L", 0x10000},
									{0x04417053, "PIC32MX764F128L", 0x10000},
									{0x04D07053, "PIC32MX130F064B", 0xF000},
									{0x04D09053, "PIC32MX130F064C", 0xF000},
									{0x04D0B053, "PIC32MX130F064D", 0xF000},
									{0x04D01053, "PIC32MX230F064B", 0xF000},
									{0x04D03053, "PIC32MX230F064C", 0xF000},
									{0x04D05053, "PIC32MX230F064D", 0xF000},
									{0x04D06053, "PIC32MX150F128B", 0x10000},
									{0x04D08053, "PIC32MX150F128C", 0x10000},
									{0x04D0A053, "PIC32MX150F128D", 0x10000},
									{0x04D00053, "PIC32MX250F128B", 0x10000},
									{0x04D02053, "PIC32MX250F128C", 0x10000},
									{0x04D04053, "PIC32MX250F128D", 0x10000},
									{0x06610053, "PIC32MX170F256B", 0x20000},
									{0x0661A053, "PIC32MX170F256D", 0x20000},
									{0x06600053, "PIC32MX270F256B", 0x20000},
									{0x0660A053, "PIC32MX270F256D", 0x20000},
									{0x0660C053, "PIC32MX270F256DB", 0x20000},
									{0x06703053, "PIC32MX130F256B", 0x20000},
									{0x06705053, "PIC32MX130F256D", 0x20000},
									{0x06700053, "PIC32MX230F256B", 0x20000},
									{0x06702053, "PIC32MX230F256D", 0x20000},
									{0x05600053, "PIC32MX330F064H", 0xF000},
									{0x05601053, "PIC32MX330F064L", 0xF000},
									{0x05704053, "PIC32MX350F256H", 0x20000},
									{0x05705053, "PIC32MX350F256L", 0x20000},
									{0x05602053, "PIC32MX430F064H", 0xF000},
									{0x05603053, "PIC32MX430F064L", 0xF000},
									{0x05706053, "PIC32MX450F256H", 0x20000},
									{0x05707053, "PIC32MX450F256L", 0x20000},
									{0x0570C053, "PIC32MX350F128H", 0x10000},
									{0x0570D053, "PIC32MX350F128L", 0x10000},
									{0x0570E053, "PIC32MX450F128H", 0x10000},
									{0x0570F053, "PIC32MX450F128L", 0x10000},
									{0x05808053, "PIC32MX370F512H", 0x40000},
									{0x05809053, "PIC32MX370F512L", 0x40000},
									{0x0580A053, "PIC32MX470F512H", 0x40000},
									{0x0580B053, "PIC32MX470F512L", 0x40000},
									{0x05710053, "PIC32MX450F256HB", 0x20000},
									{0x05811053, "PIC32MX470F512LB", 0x40000},
									{0x05103053, "PIC32MZ1024ECG064", 0x80000},
									{0x05108053, "PIC32MZ1024ECH064", 0x80000},
									{0x05130053, "PIC32MZ1024ECM064", 0x80000},
									{0x05104053, "PIC32MZ2048ECG064", 0x100000},
									{0x05109053, "PIC32MZ2048ECH064", 0x100000},
									{0x05131053, "PIC32MZ2048ECM064", 0x100000},
									{0x0510D053, "PIC32MZ1024ECG100", 0x80000},
									{0x05112053, "PIC32MZ1024ECH100", 0x80000},
									{0x0513A053, "PIC32MZ1024ECM100", 0x80000},
									{0x0510E053, "PIC32MZ2048ECG100", 0x100000},
									{0x05113053, "PIC32MZ2048ECH100", 0x100000},
									{0x0513B053, "PIC32MZ2048ECM100", 0x100000},
									{0x05117053, "PIC32MZ1024ECG124", 0x80000},
									{0x0511C053, "PIC32MZ1024ECH124", 0x80000},
									{0x05144053, "PIC32MZ1024ECM124", 0x80000},
									{0x05118053, "PIC32MZ2048ECG124", 0x100000},
									{0x0511D053, "PIC32MZ2048ECH124", 0x100000},
									{0x05145053, "PIC32MZ2048ECM124", 0x100000},
									{0x05121053, "PIC32MZ1024ECG144", 0x80000},
									{0x05126053, "PIC32MZ1024ECH144", 0x80000},
									{0x0514E053, "PIC32MZ1024ECM144", 0x80000},
									{0x05122053, "PIC32MZ2048ECG144", 0x100000},
									{0x05127053, "PIC32MZ2048ECH144", 0x100000},
									{0x0514F053, "PIC32MZ2048ECM144", 0x100000},
									{0x06A10053, "PIC32MX150F256H", 0x20000},
									{0x06A11053, "PIC32MX150F256L", 0x20000},
									{0x06A30053, "PIC32MX170F512H", 0x40000},
									{0x06A31053, "PIC32MX170F512L", 0x40000},
									{0x06A12053, "PIC32MX250F256H", 0x20000},
									{0x06A13053, "PIC32MX250F256L", 0x20000},
									{0x06A32053, "PIC32MX270F512H", 0x40000},
									{0x06A33053, "PIC32MX270F512L", 0x40000},
									{0x06A14053, "PIC32MX550F256H", 0x20000},
									{0x06A15053, "PIC32MX550F256L", 0x20000},
									{0x06A34053, "PIC32MX570F512H", 0x40000},
									{0x06A35053, "PIC32MX570F512L", 0x40000},
									{0x06A50053, "PIC32MX120F064H", 0xF000},
									{0x06A00053, "PIC32MX130F128H", 0x10000},
									{0x06A01053, "PIC32MX130F128L", 0x10000},
									{0x06A02053, "PIC32MX230F128H", 0x10000},
									{0x06A03053, "PIC32MX230F128L", 0x10000},
									{0x06A04053, "PIC32MX530F128H", 0x10000},
									{0x06A05053, "PIC32MX530F128L", 0x10000},
									{0x07201053, "PIC32MZ0512EFE064", 0x40000},
									{0x07206053, "PIC32MZ0512EFF064", 0x40000},
									{0x0722E053, "PIC32MZ0512EFK064", 0x40000},
									{0x07202053, "PIC32MZ1024EFE064", 0x80000},
									{0x07207053, "PIC32MZ1024EFF064", 0x80000},
									{0x0722F053, "PIC32MZ1024EFK064", 0x80000},
									{0x07203053, "PIC32MZ1024EFG064", 0x80000},
									{0x07208053, "PIC32MZ1024EFH064", 0x80000},
									{0x07230053, "PIC32MZ1024EFM064", 0x80000},
									{0x07204053, "PIC32MZ2048EFG064", 0x100000},
									{0x07209053, "PIC32MZ2048EFH064", 0x100000},
									{0x07231053, "PIC32MZ2048EFM064", 0x100000},
									{0x0720B053, "PIC32MZ0512EFE100", 0x40000},
									{0x07210053, "PIC32MZ0512EFF100", 0x40000},
									{0x07238053, "PIC32MZ0512EFK100", 0x40000},
									{0x0720C053, "PIC32MZ1024EFE100", 0x80000},
									{0x07211053, "PIC32MZ1024EFF100", 0x80000},
									{0x07239053, "PIC32MZ1024EFK100", 0x80000},
									{0x0720D053, "PIC32MZ1024EFG100", 0x80000},
									{0x07212053, "PIC32MZ1024EFH100", 0x80000},
									{0x0723A053, "PIC32MZ1024EFM100", 0x80000},
									{0x0720E053, "PIC32MZ2048EFG100", 0x100000},
									{0x07213053, "PIC32MZ2048EFH100", 0x100000},
									{0x0723B053, "PIC32MZ2048EFM100", 0x100000},
									{0x07215053, "PIC32MZ0512EFE124", 0x40000},
									{0x0721A053, "PIC32MZ0512EFF124", 0x40000},
									{0x07242053, "PIC32MZ0512EFK124", 0x40000},
									{0x07216053, "PIC32MZ1024EFE124", 0x80000},
									{0x0721B053, "PIC32MZ1024EFF124", 0x80000},
									{0x07243053, "PIC32MZ1024EFK124", 0x80000},
									{0x07217053, "PIC32MZ1024EFG124", 0x80000},
									{0x0721C053, "PIC32MZ1024EFH124", 0x80000},
									{0x07244053, "PIC32MZ1024EFM124", 0x80000},
									{0x07218053, "PIC32MZ2048EFG124", 0x100000},
									{0x0721D053, "PIC32MZ2048EFH124", 0x100000},
									{0x07245053, "PIC32MZ2048EFM124", 0x100000},
									{0x0721F053, "PIC32MZ0512EFE144", 0x40000},
									{0x07224053, "PIC32MZ0512EFF144", 0x40000},
									{0x0724C053, "PIC32MZ0512EFK144", 0x40000},
									{0x07220053, "PIC32MZ1024EFE144", 0x80000},
									{0x07225053, "PIC32MZ1024EFF144", 0x80000},
									{0x0724D053, "PIC32MZ1024EFK144", 0x80000},
									{0x07221053, "PIC32MZ1024EFG144", 0x80000},
									{0x07226053, "PIC32MZ1024EFH144", 0x80000},
									{0x0724E053, "PIC32MZ1024EFM144", 0x80000},
									{0x07222053, "PIC32MZ2048EFG144", 0x100000},
									{0x07227053, "PIC32MZ2048EFH144", 0x100000},
									{0x0724F053, "PIC32MZ2048EFM144", 0x100000},
									{0x05F0F053, "PIC32MZ1064DAA169", 0xF000},
									{0x05F10053, "PIC32MZ1064DAB169", 0xF000},
									{0x05F18053, "PIC32MZ2064DAA169", 0xF000},
									{0x05F19053, "PIC32MZ2064DAB169", 0xF000},
									{0x05F45053, "PIC32MZ1064DAG169", 0xF000},
									{0x05F46053, "PIC32MZ1064DAH169", 0xF000},
									{0x05F4E053, "PIC32MZ2064DAG169", 0xF000},
									{0x05F4F053, "PIC32MZ2064DAH169", 0xF000},
									{0x05F7B053, "PIC32MZ1064DAA176", 0xF000},
									{0x05F7C053, "PIC32MZ1064DAB176", 0xF000},
									{0x05F84053, "PIC32MZ2064DAA176", 0xF000},
									{0x05F85053, "PIC32MZ2064DAB176", 0xF000},
									{0x05FB1053, "PIC32MZ1064DAG176", 0xF000},
									{0x05FB2053, "PIC32MZ1064DAH176", 0xF000},
									{0x05FBA053, "PIC32MZ2064DAG176", 0xF000},
									{0x05FBB053, "PIC32MZ2064DAH176", 0xF000},
									{0x05F60053, "PIC32MZ1064DAA288", 0xF000},
									{0x05F61053, "PIC32MZ1064DAB288", 0xF000},
									{0x05F69053, "PIC32MZ2064DAA288", 0xF000},
									{0x05F6A053, "PIC32MZ2064DAB288", 0xF000},
									{0x07800053, "PIC32MX154F128B", 0x10000},
									{0x07804053, "PIC32MX154F128D", 0x10000},
									{0x07808053, "PIC32MX155F128B", 0x10000},
									{0x0780C053, "PIC32MX155F128D", 0x10000},
									{0x07801053, "PIC32MX174F256B", 0x20000},
									{0x07805053, "PIC32MX174F256D", 0x20000},
									{0x07809053, "PIC32MX175F256B", 0x20000},
									{0x0780D053, "PIC32MX175F256D", 0x20000},
									{0x07802053, "PIC32MX254F128B", 0x10000},
									{0x07806053, "PIC32MX254F128D", 0x10000},
									{0x0780A053, "PIC32MX255F128B", 0x10000},
									{0x0780E053, "PIC32MX255F128D", 0x10000},
									{0x07803053, "PIC32MX274F256B", 0x20000},
									{0x07807053, "PIC32MX274F256D", 0x20000},
									{0x0780B053, "PIC32MX275F256B", 0x20000},
									{0x0780F053, "PIC32MX275F256D", 0x20000},
									{0x06211053, "PIC32MK0512GPD064", 0x40000},
									{0x0620E053, "PIC32MK1024GPD064", 0x80000},
									{0x06210053, "PIC32MK0512GPD100", 0x40000},
									{0x0620D053, "PIC32MK1024GPD100", 0x80000},
									{0x0620B053, "PIC32MK0512GPE064", 0x40000},
									{0x06208053, "PIC32MK1024GPE064", 0x80000},
									{0x0620A053, "PIC32MK0512GPE100", 0x40000},
									{0x06207053, "PIC32MK1024GPE100", 0x80000},
									{0x06205053, "PIC32MK0512MCF064", 0x40000},
									{0x06202053, "PIC32MK1024MCF064", 0x80000},
									{0x06201053, "PIC32MK0512MCF100", 0x40000},
									{0x06201053, "PIC32MK1024MCF100", 0x80000}};
};
